@@ -1,28 +1,39 @@
 """
 local path implementation.
 """
+from contextlib import contextmanager
 import sys, os, stat, re, atexit
 import py
 from py._path import common
+from stat import S_ISLNK, S_ISDIR, S_ISREG
+
+from os.path import normpath, isabs, exists, isdir, isfile, islink
 
 iswin32 = sys.platform == "win32" or (getattr(os, '_name', False) == 'nt')
 
+if sys.version_info > (3,0):
+    def map_as_list(func, iter):
+        return list(map(func, iter))
+else:
+    map_as_list = map
+
 class Stat(object):
     def __getattr__(self, name):
         return getattr(self._osstatresult, "st_" + name)
 
-    def __init__(self, path, osstatresult): 
-        self.path = path 
+    def __init__(self, path, osstatresult):
+        self.path = path
         self._osstatresult = osstatresult
 
+    @property
     def owner(self):
         if iswin32:
             raise NotImplementedError("XXX win32")
-        import pwd 
+        import pwd
         entry = py.error.checked_call(pwd.getpwuid, self.uid)
         return entry[0]
-    owner = property(owner, None, None, "owner of path") 
 
+    @property
     def group(self):
         """ return group name of file. """
         if iswin32:
@@ -30,7 +41,16 @@ class Stat(object):
         import grp
         entry = py.error.checked_call(grp.getgrgid, self.gid)
         return entry[0]
-    group = property(group) 
+
+    def isdir(self):
+        return S_ISDIR(self._osstatresult.st_mode)
+
+    def isfile(self):
+        return S_ISREG(self._osstatresult.st_mode)
+
+    def islink(self):
+        st = self.path.lstat()
+        return S_ISLNK(self._osstatresult.st_mode)
 
 class PosixPath(common.PathBase):
     def chown(self, user, group, rec=0):
@@ -42,7 +62,7 @@ class PosixPath(common.PathBase):
         uid = getuserid(user)
         gid = getgroupid(group)
         if rec:
-            for x in self.visit(rec=lambda x: x.check(link=0)): 
+            for x in self.visit(rec=lambda x: x.check(link=0)):
                 if x.check(link=0):
                     py.error.checked_call(os.chown, str(x), uid, gid)
         py.error.checked_call(os.chown, str(self), uid, gid)
@@ -68,10 +88,6 @@ class PosixPath(common.PathBase):
             target = self.sep.join(('..', )*n + (relsource, ))
             py.error.checked_call(os.symlink, target, self.strpath)
 
-    def samefile(self, other):
-        """ return True if other refers to the same stat object as self. """
-        return py.error.checked_call(os.path.samefile, str(self), str(other))
-
 def getuserid(user):
     import pwd
     if not isinstance(user, int):
@@ -87,12 +103,12 @@ def getgroupid(group):
 FSBase = not iswin32 and PosixPath or common.PathBase
 
 class LocalPath(FSBase):
-    """ object oriented interface to os.path and other local filesystem 
-        related information. 
+    """ object oriented interface to os.path and other local filesystem
+        related information.
     """
     class ImportMismatchError(ImportError):
         """ raised on pyimport() if there is a mismatch of __file__'s"""
-        
+
     sep = os.sep
     class Checkers(common.Checkers):
         def _stat(self):
@@ -106,19 +122,19 @@ class LocalPath(FSBase):
                 return self._statcache
 
         def dir(self):
-            return stat.S_ISDIR(self._stat().mode)
+            return S_ISDIR(self._stat().mode)
 
         def file(self):
-            return stat.S_ISREG(self._stat().mode)
+            return S_ISREG(self._stat().mode)
 
         def exists(self):
             return self._stat()
 
         def link(self):
             st = self.path.lstat()
-            return stat.S_ISLNK(st.mode)
+            return S_ISLNK(st.mode)
 
-    def __new__(cls, path=None):
+    def __init__(self, path=None):
         """ Initialize and return a local Path instance.
 
         Path can be relative to the current directory.
@@ -127,21 +143,16 @@ class LocalPath(FSBase):
         Note also that passing in a local path object will simply return
         the exact same path object. Use new() to get a new copy.
         """
-        if isinstance(path, common.PathBase):
-            if path.__class__ == cls:
-                return path
-            path = path.strpath
-        # initialize the path
-        self = object.__new__(cls)
-        if not path:
-            self.strpath = os.getcwd()
+        if path is None:
+            self.strpath = py.error.checked_call(os.getcwd)
+        elif isinstance(path, common.PathBase):
+            self.strpath = path.strpath
         elif isinstance(path, py.builtin._basestring):
-            self.strpath = os.path.abspath(os.path.normpath(str(path)))
+            self.strpath = os.path.abspath(normpath(str(path)))
         else:
             raise ValueError("can only pass None, Path instances "
                              "or non-empty strings to LocalPath")
         assert isinstance(self.strpath, str)
-        return self
 
     def __hash__(self):
         return hash(self.strpath)
@@ -149,7 +160,7 @@ class LocalPath(FSBase):
     def __eq__(self, other):
         s1 = str(self)
         s2 = str(other)
-        if iswin32: 
+        if iswin32:
             s1 = s1.lower()
             s2 = s2.lower()
         return s1 == s2
@@ -160,22 +171,37 @@ class LocalPath(FSBase):
     def __lt__(self, other):
         return str(self) < str(other)
 
+    def __gt__(self, other):
+        return str(self) > str(other)
+
+    def samefile(self, other):
+        """ return True if 'other' references the same file as 'self'.
+        """
+        if not isinstance(other, py.path.local):
+            other = os.path.abspath(str(other))
+        if self == other:
+            return True
+        if iswin32:
+            return False # ther is no samefile
+        return py.error.checked_call(
+                os.path.samefile, str(self), str(other))
+
     def remove(self, rec=1, ignore_errors=False):
-        """ remove a file or directory (or a directory tree if rec=1). 
-        if ignore_errors is True, errors while removing directories will 
+        """ remove a file or directory (or a directory tree if rec=1).
+        if ignore_errors is True, errors while removing directories will
         be ignored.
         """
         if self.check(dir=1, link=0):
             if rec:
-                # force remove of readonly files on windows 
-                if iswin32: 
+                # force remove of readonly files on windows
+                if iswin32:
                     self.chmod(448, rec=1) # octcal 0700
                 py.error.checked_call(py.std.shutil.rmtree, self.strpath,
                     ignore_errors=ignore_errors)
             else:
                 py.error.checked_call(os.rmdir, self.strpath)
         else:
-            if iswin32: 
+            if iswin32:
                 self.chmod(448) # octcal 0700
             py.error.checked_call(os.remove, self.strpath)
 
@@ -197,22 +223,25 @@ class LocalPath(FSBase):
                 buf = f.read(chunksize)
                 if not buf:
                     return hash.hexdigest()
-                hash.update(buf) 
+                hash.update(buf)
         finally:
             f.close()
 
     def new(self, **kw):
         """ create a modified version of this path.
-            the following keyword arguments modify various path parts:
+            the following keyword arguments modify various path parts::
 
               a:/some/path/to/a/file.ext
-              ||                            drive
-                |-------------|             dirname
-                                |------|    basename
-                                |--|        purebasename
-                                    |--|    ext
+              xx                           drive
+              xxxxxxxxxxxxxxxxx            dirname
+                                xxxxxxxx   basename
+                                xxxx       purebasename
+                                     xxx   ext
         """
         obj = object.__new__(self.__class__)
+        if not kw:
+            obj.strpath = self.strpath
+            return obj
         drive, dirname, basename, purebasename,ext = self._getbyspec(
              "drive,dirname,basename,purebasename,ext")
         if 'basename' in kw:
@@ -229,24 +258,17 @@ class LocalPath(FSBase):
                     ext = '.' + ext
             kw['basename'] = pb + ext
 
-        kw.setdefault('drive', drive)
-        kw.setdefault('dirname', dirname)
+        if ('dirname' in kw and not kw['dirname']):
+            kw['dirname'] = drive
+        else:
+            kw.setdefault('dirname', dirname)
         kw.setdefault('sep', self.sep)
-        obj.strpath = os.path.normpath(
-            "%(drive)s%(dirname)s%(sep)s%(basename)s" % kw)
+        obj.strpath = normpath(
+            "%(dirname)s%(sep)s%(basename)s" % kw)
         return obj
-    
+
     def _getbyspec(self, spec):
-        """ return a sequence of specified path parts.  'spec' is
-            a comma separated string containing path part names.
-            according to the following convention:
-            a:/some/path/to/a/file.ext
-            ||                            drive
-              |-------------|             dirname
-                              |------|    basename
-                              |--|        purebasename
-                                  |--|    ext
-        """
+        """ see new for what 'spec' can be. """
         res = []
         parts = self.strpath.split(self.sep)
 
@@ -256,7 +278,7 @@ class LocalPath(FSBase):
             if name == 'drive':
                 append(parts[0])
             elif name == 'dirname':
-                append(self.sep.join(['']+parts[1:-1]))
+                append(self.sep.join(parts[:-1]))
             else:
                 basename = parts[-1]
                 if name == 'basename':
@@ -280,48 +302,78 @@ class LocalPath(FSBase):
         components.  if abs=1 is used restart from root if any
         of the args is an absolute path.
         """
-        if not args:
-            return self
-        strpath = self.strpath
         sep = self.sep
-        strargs = [str(x) for x in args]
-        if kwargs.get('abs', 0):
-            for i in range(len(strargs)-1, -1, -1):
-                if os.path.isabs(strargs[i]):
-                    strpath = strargs[i]
-                    strargs = strargs[i+1:]
+        strargs = map_as_list(str, args)
+        strpath = self.strpath
+        if kwargs.get('abs'):
+            newargs = []
+            for arg in reversed(strargs):
+                if isabs(arg):
+                    strpath = arg
+                    strargs = newargs
                     break
+                newargs.insert(0, arg)
         for arg in strargs:
             arg = arg.strip(sep)
             if iswin32:
                 # allow unix style paths even on windows.
                 arg = arg.strip('/')
                 arg = arg.replace('/', sep)
-            if arg:
-                if not strpath.endswith(sep):
-                    strpath += sep
-                strpath += arg
-        obj = self.new()
-        obj.strpath = os.path.normpath(strpath)
+            strpath = strpath + sep + arg
+        obj = object.__new__(self.__class__)
+        obj.strpath = normpath(strpath)
         return obj
 
-    def open(self, mode='r'):
-        """ return an opened file with the given mode. """
+    def open(self, mode='r', ensure=False):
+        """ return an opened file with the given mode.
+
+        If ensure is True, create parent directories if needed.
+        """
+        if ensure:
+            self.dirpath().ensure(dir=1)
         return py.error.checked_call(open, self.strpath, mode)
 
+    def _fastjoin(self, name):
+        child = object.__new__(self.__class__)
+        child.strpath = self.strpath + self.sep + name
+        return child
+
+    def islink(self):
+        return islink(self.strpath)
+
+    def check(self, **kw):
+        if not kw:
+            return exists(self.strpath)
+        if len(kw) == 1:
+            if "dir" in kw:
+                return not kw["dir"] ^ isdir(self.strpath)
+            if "file" in kw:
+                return not kw["file"] ^ isfile(self.strpath)
+        return super(LocalPath, self).check(**kw)
+
+    _patternchars = set("*?[" + os.path.sep)
     def listdir(self, fil=None, sort=None):
         """ list directory contents, possibly filter by the given fil func
             and possibly sorted.
         """
+        if fil is None and sort is None:
+            names = py.error.checked_call(os.listdir, self.strpath)
+            return map_as_list(self._fastjoin, names)
         if isinstance(fil, str):
+            if not self._patternchars.intersection(fil):
+                child = self._fastjoin(fil)
+                if os.path.exists(child.strpath):
+                    return [child]
+                return []
             fil = common.FNMatcher(fil)
+        names = py.error.checked_call(os.listdir, self.strpath)
         res = []
-        for name in py.error.checked_call(os.listdir, self.strpath):
-            childurl = self.join(name)
-            if fil is None or fil(childurl):
-                res.append(childurl)
+        for name in names:
+            child = self._fastjoin(name)
+            if fil is None or fil(child):
+                res.append(child)
         self._sortlist(res, sort)
-        return res 
+        return res
 
     def size(self):
         """ return size of the underlying file object """
@@ -331,14 +383,15 @@ class LocalPath(FSBase):
         """ return last modification time of the path. """
         return self.stat().mtime
 
-    def copy(self, target, archive=False):
+    def copy(self, target, mode=False):
         """ copy path to target."""
-        assert not archive, "XXX archive-mode not supported"
         if self.check(file=1):
             if target.check(dir=1):
                 target = target.join(self.basename)
             assert self!=target
             copychunked(self, target)
+            if mode:
+                copymode(self, target)
         else:
             def rec(p):
                 return p.check(link=0)
@@ -348,10 +401,13 @@ class LocalPath(FSBase):
                 newx.dirpath().ensure(dir=1)
                 if x.check(link=1):
                     newx.mksymlinkto(x.readlink())
+                    continue
                 elif x.check(file=1):
                     copychunked(x, newx)
                 elif x.check(dir=1):
                     newx.ensure(dir=1)
+                if mode:
+                    copymode(x, newx)
 
     def rename(self, target):
         """ rename this path to target. """
@@ -371,8 +427,12 @@ class LocalPath(FSBase):
         py.error.checked_call(os.mkdir, str(p))
         return p
 
-    def write(self, data, mode='w'):
-        """ write data into path. """
+    def write(self, data, mode='w', ensure=False):
+        """ write data into path.   If ensure is True create
+        missing parent directories.
+        """
+        if ensure:
+            self.dirpath().ensure(dir=1)
         if 'b' in mode:
             if not py.builtin._isbytes(data):
                 raise ValueError("can only process bytes")
@@ -418,9 +478,16 @@ class LocalPath(FSBase):
                 p.open('w').close()
             return p
 
-    def stat(self):
+    def stat(self, raising=True):
         """ Return an os.stat() tuple. """
-        return Stat(self, py.error.checked_call(os.stat, self.strpath))
+        if raising == True:
+            return Stat(self, py.error.checked_call(os.stat, self.strpath))
+        try:
+            return Stat(self, os.stat(self.strpath))
+        except KeyboardInterrupt:
+            raise
+        except Exception:
+            return None
 
     def lstat(self):
         """ Return an os.lstat() tuple. """
@@ -441,10 +508,25 @@ class LocalPath(FSBase):
 
     def chdir(self):
         """ change directory to self and return old current directory """
-        old = self.__class__()
+        try:
+            old = self.__class__()
+        except py.error.ENOENT:
+            old = None
         py.error.checked_call(os.chdir, self.strpath)
         return old
 
+
+    @contextmanager
+    def as_cwd(self):
+        """ return context manager which changes to current dir during the
+        managed "with" context. On __enter__ it returns the old dir.
+        """
+        old = self.chdir()
+        try:
+            yield old
+        finally:
+            old.chdir()
+
     def realpath(self):
         """ return a new path which contains no symbolic links."""
         return self.__class__(os.path.realpath(self.strpath))
@@ -461,8 +543,8 @@ class LocalPath(FSBase):
         return self.strpath
 
     def pypkgpath(self, pkgname=None):
-        """ return the path's package path by looking for the given
-            pkgname.  If pkgname is None then look for the last
+        """ return the Python package path by looking for a
+            pkgname.  If pkgname is None look for the last
             directory upwards which still contains an __init__.py
             and whose basename is python-importable.
             Return None if a pkgpath can not be determined.
@@ -516,24 +598,37 @@ class LocalPath(FSBase):
             if pkgpath is not None:
                 if ensuresyspath:
                     self._prependsyspath(pkgpath.dirpath())
-                pkg = __import__(pkgpath.basename, None, None, [])
+                __import__(pkgpath.basename)
+                pkg = sys.modules[pkgpath.basename]
                 names = self.new(ext='').relto(pkgpath.dirpath())
                 names = names.split(self.sep)
+                if names and names[-1] == "__init__":
+                    names.pop()
                 modname = ".".join(names)
             else:
                 # no package scope, still make it possible
                 if ensuresyspath:
                     self._prependsyspath(self.dirpath())
                 modname = self.purebasename
-            mod = __import__(modname, None, None, ['__doc__'])
+            __import__(modname)
+            mod = sys.modules[modname]
+            if self.basename == "__init__.py":
+                return mod # we don't check anything as we might
+                       # we in a namespace package ... too icky to check
             modfile = mod.__file__
             if modfile[-4:] in ('.pyc', '.pyo'):
                 modfile = modfile[:-1]
             elif modfile.endswith('$py.class'):
                 modfile = modfile[:-9] + '.py'
-            if modfile.endswith("__init__.py"):
-                modfile = modfile[:-12]
-            if not self.samefile(modfile):
+            if modfile.endswith(os.path.sep + "__init__.py"):
+                if self.basename != "__init__.py":
+                    modfile = modfile[:-12]
+
+            try:
+                issame = self.samefile(modfile)
+            except py.error.ENOENT:
+                issame = False
+            if not issame:
                 raise self.ImportMismatchError(modname, modfile, self)
             return mod
         else:
@@ -553,13 +648,13 @@ class LocalPath(FSBase):
 
     def sysexec(self, *argv, **popen_opts):
         """ return stdout text from executing a system child process,
-            where the 'self' path points to executable. 
-            The process is directly invoked and not through a system shell. 
+            where the 'self' path points to executable.
+            The process is directly invoked and not through a system shell.
         """
         from subprocess import Popen, PIPE
-        argv = map(str, argv)
+        argv = map_as_list(str, argv)
         popen_opts['stdout'] = popen_opts['stderr'] = PIPE
-        proc = Popen([str(self)] + list(argv), **popen_opts)
+        proc = Popen([str(self)] + argv, **popen_opts)
         stdout, stderr = proc.communicate()
         ret = proc.wait()
         if py.builtin._isbytes(stdout):
@@ -571,7 +666,7 @@ class LocalPath(FSBase):
                                            stdout, stderr,)
         return stdout
 
-    def sysfind(cls, name, checker=None):
+    def sysfind(cls, name, checker=None, paths=None):
         """ return a path object found by looking at the systems
             underlying PATH specification. If the checker is not None
             it will be invoked to filter matching paths.  If a binary
@@ -584,21 +679,23 @@ class LocalPath(FSBase):
             if p.check(file=1):
                 return p
         else:
-            if iswin32:
-                paths = py.std.os.environ['Path'].split(';')
-                if '' not in paths and '.' not in paths:
-                    paths.append('.')
-                try:
-                    systemroot = os.environ['SYSTEMROOT']
-                except KeyError:
-                    pass
+            if paths is None:
+                if iswin32:
+                    paths = py.std.os.environ['Path'].split(';')
+                    if '' not in paths and '.' not in paths:
+                        paths.append('.')
+                    try:
+                        systemroot = os.environ['SYSTEMROOT']
+                    except KeyError:
+                        pass
+                    else:
+                        paths = [re.sub('%SystemRoot%', systemroot, path)
+                                 for path in paths]
                 else:
-                    paths = [re.sub('%SystemRoot%', systemroot, path)
-                             for path in paths]
-                tryadd = '', '.exe', '.com', '.bat' # XXX add more?
-            else:
-                paths = py.std.os.environ['PATH'].split(':')
-                tryadd = ('',)
+                    paths = py.std.os.environ['PATH'].split(':')
+            tryadd = ['']
+            if iswin32:
+                tryadd += os.environ['PATHEXT'].split(os.pathsep)
 
             for x in paths:
                 for addext in tryadd:
@@ -618,7 +715,10 @@ class LocalPath(FSBase):
         try:
             x = os.environ['HOME']
         except KeyError:
-            x = os.environ["HOMEDRIVE"] + os.environ['HOMEPATH']
+            try:
+                x = os.environ["HOMEDRIVE"] + os.environ['HOMEPATH']
+            except KeyError:
+                return None
         return cls(x)
     _gethomedir = classmethod(_gethomedir)
 
@@ -725,7 +825,7 @@ class LocalPath(FSBase):
                         raise
                     except: # this might be py.error.Error, WindowsError ...
                         pass
-        
+
         # make link...
         try:
             username = os.environ['USER']           #linux, et al
@@ -743,12 +843,15 @@ class LocalPath(FSBase):
             pass
         try:
             os.symlink(src, dest)
-        except (OSError, AttributeError): # AttributeError on win32
+        except (OSError, AttributeError, NotImplementedError):
             pass
 
         return udir
     make_numbered_dir = classmethod(make_numbered_dir)
 
+def copymode(src, dest):
+    py.std.shutil.copymode(str(src), str(dest))
+
 def copychunked(src, dest):
     chunksize = 524288 # half a meg of bytes
     fsrc = src.open('rb')
@@ -765,41 +868,6 @@ def copychunked(src, dest):
     finally:
         fsrc.close()
 
-def autopath(globs=None):
-    """ (deprecated) return the (local) path of the "current" file pointed to by globals or - if it is none - alternatively the callers frame globals.
-
-        the path will always point to a .py file  or to None.
-        the path will have the following payload:
-        pkgdir   is the last parent directory path containing __init__.py 
-    """
-    py.log._apiwarn("1.1", "py.magic.autopath deprecated, "
-        "use py.path.local(__file__) and maybe pypkgpath/pyimport().")
-    if globs is None:
-        globs = sys._getframe(1).f_globals
-    try:
-        __file__ = globs['__file__']
-    except KeyError:
-        if not sys.argv[0]:
-            raise ValueError("cannot compute autopath in interactive mode")
-        __file__ = os.path.abspath(sys.argv[0])
-
-    ret = py.path.local(__file__)
-    if ret.ext in ('.pyc', '.pyo'):
-        ret = ret.new(ext='.py')
-    current = pkgdir = ret.dirpath()
-    while 1:
-        if current.join('__init__.py').check():
-            pkgdir = current
-            current = current.dirpath()
-            if pkgdir != current:
-                continue
-        elif str(current) not in sys.path:
-            sys.path.insert(0, str(current))
-        break
-    ret.pkgdir = pkgdir
-    return ret
-
-
 def isimportable(name):
     if name:
         if not (name[0].isalpha() or name[0] == '_'):
